from __future__ import annotations

import argparse
import queue
import shutil
import socket
import sys
import threading
import time
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsWrite


def info(msg: str) -> None:
    print(msg, file=sys.stderr)


def copier(
    fsrc: SupportsRead[bytes], fdst: SupportsWrite[bytes], exc_q: queue.SimpleQueue[BaseException | None], quit_after: float | None = None
) -> None:
    try:
        shutil.copyfileobj(fsrc, fdst)
    except BaseException as e:
        exc_q.put(e)
        return

    if not quit_after:
        exc_q.put(None)
        return

    time.sleep(quit_after)
    exc_q.put(TimeoutError())


def main() -> None:
    p = argparse.ArgumentParser()
    p.add_argument('-v', '--verbose', action='store_true')
    p.add_argument('-l', '--listen', action='store_true')
    p.add_argument('-p', '--local-port', type=int)
    p.add_argument('-w', '--idle-timeout', type=float)
    p.add_argument('-q', '--eof-quit-time', type=float)
    p.add_argument('host', nargs='?')
    p.add_argument('port', type=int, nargs='?')
    args = p.parse_args()

    verb = info if args.verbose else lambda _: None

    s = socket.socket()

    if args.listen:
        if not args.local_port:
            p.error('Must specify local port to listen to')
        if socket.has_dualstack_ipv6():
            srv = socket.create_server(('', args.local_port), family=socket.AF_INET6, dualstack_ipv6=True)
        else:
            srv = socket.create_server(('', args.local_port))
        verb(f'Listening on port {s.getsockname()[1]}')
        sock, client_addr = srv.accept()
        verb(f'Connection from {client_addr[0]}:{client_addr[1]}')
    else:
        if not (args.host and args.port):
            p.error('Must specify host and port to connect to')
        verb(f'Connecting to {args.host}:{args.port}')
        try:
            sock = socket.create_connection((args.host, args.port), source_address=('', args.local_port) if args.local_port else None)
        except Exception as e:
            info(f'Error! {e}')
            raise SystemExit(1) from None

    sock.settimeout(args.idle_timeout)

    exc_q = queue.SimpleQueue[BaseException | None]()
    worker_threads = [
        threading.Thread(target=copier, args=(open(0, 'rb', buffering=0), sock.makefile('wb', buffering=0), exc_q, args.eof_quit_time), daemon=True),
        threading.Thread(target=copier, args=(sock.makefile('rb', buffering=0), open(1, 'wb', buffering=0), exc_q), daemon=True),
    ]
    for thread in worker_threads:
        thread.start()
    try:
        if (exc := exc_q.get()) is not None:
            raise exc
        verb('Done')
    except (KeyboardInterrupt, BrokenPipeError, TimeoutError):
        pass
    except ConnectionError as e:
        verb(str(e))


if __name__ == '__main__':
    main()
